'''
Features:
+ Provides motion equations for multibody dynamics from 
given masses, positions, injected forces and so on.

Todo:

Change logs:
2009/05/10 v0.1
+ Rewrote the code in Python + Sympy from MuPAD.
+ Created the code for unit test, and confirmed it returns 
the correct equations for Newton's second law and Euler's 
equations (rigid body dynamics).
'''

import sympy
import unittest

def get_withKane(t, u, m_c, p_c, I_c, w_c, p_f, f_e, w_n, n_e):
    
    results = []
    for u_i in u:
        
        terms_i = []
        for k, p_c_k in enumerate(p_c):
            
            p_c_d = p_c_k.diff(t)
            p_c_dd = p_c_d.diff(t)
            v_c_k = partialDiff(p_c_d, u_i)
            terms_p = m_c[k] * p_c_dd.dot(v_c_k)
            
            w_c_d = w_c[k].diff(t)
            w_c_k = partialDiff(w_c[k], u_i)
            terms_r = (I_c[k]*w_c_d + w_c[k].cross(I_c[k]*w_c[k]).T).dot(w_c_k)
            terms_i.append( terms_p + terms_r )
        
        terms_f = []
        for k, f_e_k in enumerate(f_e):
            
            p_f_d = p_f[k].diff(t)
            v_f_k = partialDiff(p_f_d, u_i)
            terms_f.append( f_e_k.dot(v_f_k) )
        
        terms_n = []
        for k, n_e_k in enumerate(n_e):
            
            w_n_k = partialDiff(w_n[k], u_i)
            terms_n.append( n_e_k.dot(w_n_k) )
        
        def add_all(list):
            return reduce(lambda a, b: a + b, list)
            
        results.append( sympy.Eq(
            add_all(terms_i),
            add_all(terms_f) + add_all(terms_n)) )
    
    return results

class __Test_Kane(unittest.TestCase):
    
    def testNewton2LExample(self):
        print "\n" + "-"*70
        
        mat = sympy.Matrix
        t, x, m, f = sympy.symbols(
            "t, x, m, f")
        
        pos_vec = mat([x(t),0,0])
        force_vec = mat([f(t),0,0])
        I_mat = sympy.zeros(3)
        ome_vec = sympy.zeros([3,1])
        torque_vec = sympy.zeros([3,1])
        
        result = get_withKane(
            t, [x(t).diff(t)], [m],
            [pos_vec],
            [I_mat],
            [ome_vec], 
            [pos_vec], [force_vec], 
            [ome_vec], [torque_vec])
        sympy.pprint(result)
        print """
Expected output is:
      2               
     d                
[m*-----(x(t)) = f(t)]
   dt dt
Details: https://secure.wikimedia.org/wikipedia/en/wiki/Newton%27s_laws_of_motion#Newton.27s_second_law
"""

    def testEulerEquation(self):
        print "\n" + "-"*70
        
        mat = sympy.Matrix
        t, m, I_1, I_2, I_3 = sympy.symbols(
            "t, m, I_1, I_2, I_3")
        x, y, z, ome_1, ome_2, ome_3, tau = sympy.symbols(
            "x, y, z, ome_1, ome_2, ome_3, tau")
        
        pos_vec = mat([x(t),y(t),z(t)])
        force_vec = mat([0,0,0])
        I_mat = mat([[I_1,0,0],[0,I_2,0],[0,0,I_3]])
        ome_vec = mat([ome_1(t),ome_2(t),ome_3(t)])
        torque_vec = mat([0,0,tau(t)])
        
        result = get_withKane(
            t, [ome_1(t),ome_2(t),ome_3(t)], [m],
            [pos_vec],
            [I_mat],
            [ome_vec], 
            [pos_vec], [force_vec], 
            [ome_vec], [torque_vec])
        sympy.pprint(result)
        print """
Expected output is:
     d                                                                     d  
[I_1*--(ome_1(t)) + I_3*ome_2(t)*ome_3(t) - I_2*ome_2(t)*ome_3(t) = 0, I_2*--(
     dt                                                                    dt 

                                                                   d          
ome_2(t)) + I_1*ome_1(t)*ome_3(t) - I_3*ome_1(t)*ome_3(t) = 0, I_3*--(ome_3(t)
                                                                   dt         

                                                           
) + I_2*ome_1(t)*ome_2(t) - I_1*ome_1(t)*ome_2(t) = tau(t)]
                                                           
Details: https://secure.wikimedia.org/wikipedia/en/wiki/Euler%27s_equations_(rigid_body_dynamics)
"""

def partialDiff(formula, *variables):
    """Partial difference"""
    
    x = sympy.symbols("partialdiff_variable,")
    for var in variables:
        
        formula = formula.subs(var, x)
        formula = formula.diff(x)
        formula = formula.subs(x, var)
        x = var
    
    return formula

class __Test_PD(unittest.TestCase):
    
    def test(self):
        print "\n" + "-"*70
        
        x, y = sympy.symbols("x, y")
        f = y**x - sympy.cos(x)*sympy.sin(y) + x*2 + y + 1
        sympy.pprint( partialDiff(f, x) )
        print """
Expected output is:
     x                       
2 + y *log(y) + sin(x)*sin(y)
"""